/*
			Copyright (c) 1993 by
			Advanced Visual Systems Inc.
			All Rights Reserved

	This software comprises unpublished confidential information of
	Advanced Visual Systems Inc. and may not be used, copied or made
	available to anyone, except in accordance with the license
	under which it is furnished.

	This file is under Perforce control
	$Id: //depot/express/fcs992/modules/cty_scp.c#1 $
*/
/*
 * Express City Scape Module Code
 *
 * Purpose:
 *    Take a 2D input array of node data, and create a cell set
 *    of 3D blocks
 *
 * Modification History:
 *    5 Dec 1994 - ianc  - revised from CellGen work
 *    7 Dec 1994 - ianc  - revised for 1D Nvect, N-component
 *   21 Dec 1994 - ianc  - added auto norm, 2nd input field
 *
 * The generation code is structured such that
 * each cell is added incrementally,
 * and you don't need to know how may cells you
 * are going to add ahead of time.
 *
 * This is also a memory efficient method,
 * where ARRalloc and ARRrealloc are used
 * to define a local (expandable) buffer.
 * When complete, these buffers are handed
 * to the OM as part of the output field.
 * In this way no data need be copied.
 *
 * Each cell is added by copying over data
 * from a small local buffer, for
 * application design simplicity, but this
 * could be changed without a large re-write.
 */

/*
 * ===== INCLUDES ========
 */
#include <avs/port.h>
#include <avs/dtype.h>
#include <avs/err.h>
#include <avs/om.h>
#include <avs/fld.h>

/*
 * ===== MACROS ========
 */
#define ERR_RETURN(A) \
      { ERRerror("cityscape",1,ERR_ORIG, A); return(0);}

#define MAX_NAME_SIZE  1024
#define MAX_LABEL_SIZE 2048
#define ALLOC_SIZE     1000


/*
 * Macro Utility to resolve a real arg
 */
#define RESOLVE_OBJ(elem_id,string,flag)                     \
  {local_id = OMfind_subobj((elem_id),                       \
              OMstr_to_name((string)),flag);                 \
  if ( OMis_null_obj(local_id))                              \
  {                                                          \
      ERRerror("City Scape",1,ERR_ORIG,                      \
               "search failed for subobj name");             \
      return(0);                                             \
  }}

#define RESOLVE_INT_PARM(elem_id,string,flag,ii)             \
  {local_id = OMfind_subobj((elem_id),                       \
              OMstr_to_name((string)),flag);                 \
  if ( OMis_null_obj(local_id))                              \
  {                                                          \
      ERRerror("City Scape",1,ERR_ORIG,                      \
               "search failed for subobj int param name");   \
      return(0);                                             \
  }                                                          \
  if ( OMget_int_val(local_id,&(ii)) != 1 )                  \
  {                                                          \
      ERRerror("City Scape",1,ERR_ORIG,                      \
               "search failed for subobj int param value");  \
      return(0);                                             \
  }                                                          \
  }

#define RESOLVE_FLT_PARM(elem_id,string,flag,xx)             \
  {local_id = OMfind_subobj((elem_id),                       \
              OMstr_to_name((string)),flag);                 \
  if ( OMis_null_obj(local_id))                              \
  {                                                          \
      ERRerror("City Scape",1,ERR_ORIG,                      \
               "search failed for subobj flt param name");   \
      return(0);                                             \
  }                                                          \
  if ( OMget_real_val(local_id,&(xd)) != 1 )                 \
  {                                                          \
      ERRerror("City Scape",1,ERR_ORIG,                      \
               "search failed for subobj flt param value");  \
      return(0);                                             \
  }                                                          \
  (xx) = xd;                                                 \
  }


#define RESOLVE_STR_PARM(elem_id,string,flag,cc)             \
  {local_id = OMfind_subobj((elem_id),                       \
              OMstr_to_name((string)),flag);                 \
  if ( OMis_null_obj(local_id))                              \
  {                                                          \
      ERRerror("City Scape",1,ERR_ORIG,                      \
               "search failed for subobj str param name");   \
      return(0);                                             \
  }                                                          \
  if ( OMget_str_val(local_id,(char **) &(cc),               \
		     MAX_LABEL_SIZE) != 1 )                  \
  {                                                          \
      ERRerror("City Scape",1,ERR_ORIG,                      \
               "search failed for subobj str param value");  \
      return(0);                                             \
  }                                                          \
  }



/*
 * ===== STRUCTS ========
 * temp storage in local process
 * for data, prior to posting to
 * output field.
 */
typedef struct _OUTmesh {
   /*
    * Define the Storage Area
    */
	int nnodes;             /* number of output nodes */
	float *xyz;             /* coord array */
	char *data;             /* node data storage */
	int   size;             /* size of node data array */
	int   nline;            /* number of lines */
	int   *line_conn;       /* line connectivity */
	int   line_size;        /* line array alloc buffer size */
	int   ntri;             /* number of tris */
	int   *tri_conn;        /* tri connectivity */
	int   tri_size;         /* tri array alloc buffer size */
	int   nquad;            /* number of quads */
	int   *quad_conn;       /* quad connectivity */
	int   quad_size;        /* quad array alloc buffer size */
   /*
    * Define the Requested Output Data Types
    */
	int	     nspace;   /* 1D, 2D, 3D, etc. */
	int	     comp;     /* node data component number */
	int	     dtype;    /* node data storage type */
	int	     veclen;   /* node data veclen */
	int	     data_flag;/* flag to allocate node data or not */
	int	     null_flag;/* flag to allocate node data or not */
	double       null_value; /* value of null meaning */

} OUTmesh;


/*
 * input parameter block elements
 */
typedef struct {

   /**********************
    * Global Input Parms *
    **********************/
   float zscale;          /* relative height of plot */
   float fwidth;          /* width of each block, as fraction of spacing */
   float foffset;         /* side-bar offset factor from main plot */
   float threshold;       /* data value threshold for threshold sheet */
   int   h_label_inc;     /* label per tic spacing control */
   int   draw_cityscape;  /* toggle, draw main plot or not */
   int   draw_sidebars;   /* toggle, draw summary side bars or not */
   int   draw_ticmarks;   /* toggle, draw line tics on side or not */
   int   draw_labels;     /* toggle, draw text labels or not */
   int   draw_threshold;  /* toggle, show threshold sheet or hide it */
   int   thresh_choice;   /* numeric mode control for various thresh options */
   int   hollow_mode;     /* toggle, hollow/wire mode on or off */
   int   auto_norm;       /* toggle, normalize mode on or off */

   /*
    * local vars derived from input field
    */
   float min;             /* overall data min, max */
   float max;             /* overall data min, max */

} ParamBlockStruct;


int FUNCcityscape (OMobj_id in, OMobj_id in2, ParamBlockStruct *pb,
                   OUTmesh *h, OUTmesh *h2, OMobj_id out, OMobj_id out2 );
int FUNCinit_cityscape( OUTmesh *h, OMobj_id out );
int FUNCcomp_cityscape( OUTmesh *h, OMobj_id out, char *label, char *units );

static void define_out_types( OUTmesh *h, int data_flag, int comp, int veclen,
                              int dtype, int nspace,
                              double null_value, int null_flag );
static int CSbump_nodes ( OUTmesh *h, int new );
static int CSadd_node_data ( OUTmesh *h, int new, char *ldata );
static int CSbump_line_conn ( OUTmesh *h );
static int CSbump_tri_conn ( OUTmesh *h );
static int CSbump_quad_conn ( OUTmesh *h );
static int CScreate_line ( OUTmesh *h, float *points, char *ldata );
static int CScreate_tri ( OUTmesh *h, float *points, char *qdata );
static int CScreate_quad ( OUTmesh *h, float *points, char *qdata );
static int CScreate_polyline ( OUTmesh *h, float *points, char *ldata, int numpoints );
static void CSprintoutstruct( OUTmesh *h );
static int res_params ( OMobj_id object, ParamBlockStruct *pb );
static int CScreate_box ( ParamBlockStruct *pb, OUTmesh *h,
                          float xpos, float ypos, float height, float width,
                          float data_val, int hilite );
static int draw_wire_box ( ParamBlockStruct *pb, OUTmesh *h,
                           float xpos, float ypos, float height, float width,
                           float data_val, int hilite );
static int draw_city( ParamBlockStruct *pb, OUTmesh *h, int nx, int ny,
                      float **values, float **values2,
                      int veclen, int ncomp, int h_flag );
static int draw_base_line ( ParamBlockStruct *pb, OUTmesh *h, int nx, int ny );
static int draw_threshold ( ParamBlockStruct *pb, OUTmesh *h, int nx, int ny );
static int draw_averages( ParamBlockStruct *pb, OUTmesh *h, int nx, int ny,
                          float **values, int veclen, int ncomp );
/*
 * ==================================
 * ===== PRIMARY ENTRY POINT ========
 * ==================================
 */

int DVcity_scape_update(OMobj_id elem)
{
	/*
 	 * ===== PRIVATE LOCAL DATA ========
 	 */
	OMobj_id src_id, src2_id, dst_id, dst2_id, local_id;
        ParamBlockStruct pblk, *pb;
	OUTmesh  hh, *h;
	OUTmesh  hh2, *h2;

	/*
 	 * ===== RESOLVE POINTERS to FIELDS ========
 	 */
        RESOLVE_OBJ(elem,"in",OM_OBJ_RD); src_id = local_id;
        RESOLVE_OBJ(elem,"out1",OM_OBJ_RW); dst_id = local_id;
        RESOLVE_OBJ(elem,"out2",OM_OBJ_RW); dst2_id = local_id;
        src2_id = OMfind_subobj(elem,OMstr_to_name("in2"),OM_OBJ_RD);

	/*
 	 * ===== RESOLVE POINTERS to PARAMETERS ========
 	 */
        pb = &pblk;
        h  = &hh;
        h2 = &hh2;
        if (res_params(elem,pb) != 1)
		ERR_RETURN("res_params error return status");


	/*
 	 * ===== CALL OPERATOR FUNCTION ========
 	 */
	if (FUNCcityscape(src_id, src2_id,
		       pb,h,h2,
		       dst_id,
		       dst2_id ) != 1)
		ERR_RETURN("FUNCcityscape error return status");

	return(1);
}

/*
 * ===== FUNCTION to COMPUTE STRUCTURE ========
 */
int FUNCcityscape (OMobj_id in, OMobj_id in2,
		ParamBlockStruct *pb,
		OUTmesh *h, OUTmesh *h2,
		OMobj_id out, OMobj_id out2)
{
	/*
 	 * ===== PRIVATE LOCAL DATA ========
 	 */
        OMobj_id local_id;
	int i, j, stat, size, data_id;
	int out_nsets, set;
	char  units[MAX_NAME_SIZE], label[MAX_NAME_SIZE];
	int data_flag;
	float points[100];
        int   input_nnodes, input_ncomp, input_ncomp2, input_veclen, nx, ny;
        int   input_dtype, input_nvals, input_size, type;
        int   input_ndim, *input_dims;
        float **values;
        float **values2;
	int   nvect_mode;
        float min, max, fmin, fmax;
        char *minvec;
        char *maxvec;

        /**********
	 * Begin  *
	 **********/

        /*
         * Get input field characteristics
         */
        local_id = in;
	nvect_mode = 0;
        if (FLDget_nnodes(local_id,&input_nnodes) != 1)
           ERR_RETURN("Error getting nnodes on input node data");
        if (FLDget_node_data_ncomp(local_id,&input_ncomp) != 1)
           ERR_RETURN("Error getting ncomp on input node data");
        if (input_ncomp <=0 )
           ERR_RETURN("ncomp must be at least one on input node data");
        if (FLDget_node_data_veclen(local_id,0,&input_veclen) != 1)
           ERR_RETURN("Error getting veclen of input node data");
        if (input_veclen <=0 )
           ERR_RETURN("veclen must be at least one on input node data");
        if (FLDget_ndim(local_id,&input_ndim) != 1)
           ERR_RETURN("Error getting ndim of input node data");
        if (FLDget_dims(local_id,&input_dims,&input_size) != 1)
           ERR_RETURN("Error getting dims array of input node data");

        input_dtype = DTYPE_FLOAT;
        type = DTYPE_FLOAT;

        if ( input_ncomp > 1 )
	{
	    for (i=0; i<input_ncomp; i++)
	    {
                if (FLDget_node_data_veclen(local_id,i,&j) != 1)
                       ERR_RETURN("Error getting veclen of input node data");
                if ( j != 1 )
                       ERR_RETURN("veclen of each component must be 1");
		/**
                if (FLDget_node_data_minmax_vec(local_id,i,
			(char **)&minvec, (char **)&maxvec,
			&dtype, OM_GET_ARRAY_RD) != 1)
                       ERR_RETURN("Error getting min_max");

                if ( i==0 )
		{
		  min = minvec[0];
		  max = maxvec[0];
		}
                if ( min > minvec[0] ) min = minvec[0];
                if ( max < maxvec[0] ) max = maxvec[0];
		**/
		min=0.0; max=100.0;
	    }
	}
	else
	{
            if (FLDget_node_data_minmax_vec(local_id,0,
					    (char **)&minvec, (char **)&maxvec,
					    &type, OM_GET_ARRAY_RD) != 1)
		    ERR_RETURN("Error getting min_max");

	    for (i=0; i<input_veclen; i++)
	    {
                if ( i==0 )
		{
			UTILtype_to_float(&max, (char *)maxvec, type);
			UTILtype_to_float(&min, (char *)minvec, type);
		}
		UTILtype_to_float(&fmax, (char *)(maxvec+i*DTYPEtype_size[type]), type);
		UTILtype_to_float(&fmin, (char *)(minvec+i*DTYPEtype_size[type]), type);
                if ( min > fmin ) min = fmin;
                if ( max < fmax ) max = fmax;
	    }
	}
        pb->min = min;
	pb->max = max;

        if ( input_ndim > 2 )
	{
           ERR_RETURN("input Grid_Struct must be dimension <=2");
	}
        else if ( input_ndim == 2 && input_veclen > 1 )
	{
           ERR_RETURN("2D input Grid_Struct must be scalar");
	}
        else if ( input_ndim == 1 && input_ncomp > 1 && input_veclen > 1 )
	{
           ERR_RETURN("1D Nvect input components must be scalar");
	}
        else if ( input_ndim == 2 && input_veclen == 1 )
	{
            nx = input_dims[0];
            ny = input_dims[1];
	}
	else if ( input_ndim == 1 && input_veclen > 1)
	{
            nx = input_veclen;
            ny = input_dims[0];
	    input_veclen = 1;
	}
	else if ( input_ndim == 1 && input_veclen == 1 && input_ncomp > 1 )
	{
            nx = input_dims[0];
            ny = input_ncomp;
	    input_veclen = 1;
	}
        else
           ERR_RETURN("could not assess input field characteristics");

	/*
	 * Get pointers to the input node data
	 */
	values  = (float **)malloc(sizeof(float *) * input_ncomp);

	for ( i=0; i<input_ncomp; i++)
	{
           if (FLDget_typed_node_data(local_id,i,&input_dtype,
				      (char **) &(values[i]), &input_size, OM_GET_ARRAY_RD)!=1)
               ERR_RETURN("could not get input node data array pointer");
        }
        if ( !OMis_null_obj(in2) )
	{
	    if (FLDget_node_data_ncomp(in2,&input_ncomp2) != 1)
		    ERR_RETURN("Error getting ncomp on input node data 2");
	    if (input_ncomp != input_ncomp2) {
		    ERR_RETURN("number of components in in2 shoud be the same as in in");
	    }
	    values2  = (float **)malloc(sizeof(float *) * input_ncomp2);
	    for ( i=0; i<input_ncomp2; i++)
	    {
               if (FLDget_typed_node_data(in2,i,&input_dtype,
					  (char **) &(values2[i]), &input_size, OM_GET_ARRAY_RD)!=1)
                   ERR_RETURN("could not get 2nd input node data array pointer");
            }
	}
	else
	    values2 = NULL;

        /*
	 * Set Output Field Characteristics
	 */
        define_out_types( h,           /* structure to fill */
			  1,           /* node data or not */
			  0,           /* which component */
			  1,           /* vect or scalar */
			  DTYPE_FLOAT, /* data storage type */
			  3,           /* 2D, 3D, etc. */
			  -999.0,      /* value, if present */
			  0 );         /* null control flag */

	strcpy(label, "cityscape_label");
	strcpy(units, "meters");

        /*
	 * Initialize internal data structures
	 */
        if (FUNCinit_cityscape(h,out) != 1)
		ERR_RETURN("Error in FUNCinit_cityscape");

	/********************************/
	/*   Now Populate Cell Sets     */
	/********************************/

        /* Draw DownTown Sky Scrapers */
        if ( pb->draw_cityscape )
        {
            if (draw_city( pb, h, nx, ny, values, values2,
			 input_veclen, input_ncomp, 0 ) != 1)
	                 ERR_RETURN("Error in draw_city");
        }

        /* create the side averages */
        if ( pb->draw_sidebars )
        {
            if (draw_averages( pb, h, nx, ny, values,
			 input_veclen, input_ncomp ) != 1)
	                 ERR_RETURN("Error in draw_averages");
        }

        /* create the threshold sheet */
        if ( pb->draw_threshold )
        {
            if (draw_threshold( pb, h, nx, ny ) != 1)
	                 ERR_RETURN("Error in draw_threshold");
        }


        /*
	 *******************************************
	 * Complete the Output Field structures from
	 * the template local structure.
	 *******************************************
	 */
        if ( FUNCcomp_cityscape(h, out, label, units ) != 1 )
		ERR_RETURN("Error in FUNCcomp_cityscape");

        /*
	 *******************************************
	 * Phase 2, repeat steps to send data to
	 * secondary line output field.
	 *******************************************
	 */

        /*
	 * Initialize internal data structures
	 */
        define_out_types( h2,           /* structure to fill */
			  1,           /* node data or not */
			  0,           /* which component */
			  1,           /* vect or scalar */
			  DTYPE_FLOAT, /* data storage type */
			  3,           /* 2D, 3D, etc. */
			  -999.0,      /* value, if present */
			  0 );         /* null control flag */

        if (FUNCinit_cityscape(h2,out2) != 1)
		ERR_RETURN("Error in Phase 2 FUNCinit_cityscape");

        /* Draw DownTown Sky Scrapers HOLLOW cells, wireframe */
        if ( pb->draw_cityscape )
        {
            if (draw_city( pb, h2, nx, ny, values, values2,
			 input_veclen, input_ncomp, 1 ) != 1)
	                 ERR_RETURN("Error in draw_city");
        }

        /* draw a base line around city */
        if ( pb->draw_ticmarks )
        {
            if (draw_base_line ( pb, h2, nx, ny ) != 1)
	                 ERR_RETURN("Error in draw_baseline");
        }

        /*
	 * Complete the second Output Field structures from
	 * the template local structure.
	 */
        if ( FUNCcomp_cityscape(h2, out2, label, units ) != 1 )
		ERR_RETURN("Error in Phase 2 FUNCcomp_cityscape");


	if (values) free (values);
	if (values2) free (values2);

	return(1);
}

/*
 * ======= Set Output Field Characteristics =======
 */
static void
define_out_types( OUTmesh *h,      /* structure to fill */
		  int data_flag,   /* node data or not */
		  int comp,        /* which component */
		  int veclen,      /* vect or scalar */
		  int dtype,       /* data storage type */
		  int nspace,      /* 2D, 3D, etc. */
		  double null_value,  /* value, if present */
		  int null_flag )  /* null control flag */
{
    h->data_flag     = data_flag;
    h->comp          = comp;
    h->veclen        = veclen;
    h->dtype         = dtype;
    h->nspace        = nspace;
    h->null_value    = null_value;
    h->null_flag     = null_flag;
}

/*
 * ===== INITIALIZE INTERNALS  ========
 */
int FUNCinit_cityscape( OUTmesh *h,        /* internal structure storage */
                        OMobj_id out )     /* elem_id of output field */
{

	/********************************/
	/*   Free pre-allocated arrays  */
	/********************************/
	if (FLDset_ncell_sets (out, 0) != 1) {
		ERR_RETURN("Error setting ncell_sets");
	}
	if (FLDset_nnodes (out, 0) != 1) {
		ERR_RETURN("Error setting nnodes");
	}
	h->xyz = NULL;
	h->data = NULL;
	if (FLDset_coord(out, h->xyz, 0, OM_SET_ARRAY_FREE) != 1) {
		ERR_RETURN("Error setting coord");
	}
	FLDset_node_data(out, 0, (char *)h->data, h->dtype,
			 0, OM_SET_ARRAY_FREE);

	/********************************/
	/*Initially Allocate Temp arrays*/
	/********************************/

	h->nnodes = 0;
	h->size = ALLOC_SIZE;
	h->xyz = (float *)ARRalloc(NULL,
		      DTYPE_FLOAT, h->nspace*h->size, NULL);
	if (h->data_flag)
		h->data = (char *)ARRalloc(NULL,
		      h->dtype,
                      DTYPEtype_size[h->dtype]*h->size*h->veclen,
                      NULL);
	else
		h->data = NULL;
	h->nline = 0;
	h->ntri  = 0;
	h->nquad = 0;
	h->line_size = ALLOC_SIZE;
	h->line_conn = (int *)ARRalloc(NULL, DTYPE_INT, h->line_size, NULL);
	h->tri_size  = ALLOC_SIZE;
	h->tri_conn  = (int *)ARRalloc(NULL, DTYPE_INT, h->tri_size, NULL);
	h->quad_size = ALLOC_SIZE;
	h->quad_conn = (int *)ARRalloc(NULL, DTYPE_INT, h->quad_size, NULL);

	return(1);
}

/*
 * ===== COMPLETE OUTPUT FIELD  ========
 */
int FUNCcomp_cityscape( OUTmesh *h,     /* internal structure storage */
                        OMobj_id out,   /* elem_id of output field */
		        char *label,    /* node data labels */
		        char *units )   /* coord and node data units */
{
	OMobj_id cell_set;
        int set, out_nsets;

	out_nsets = 0;
	set = 0;

	/********************************/
	/*   Transfer Temp Storage      */
	/*     Nodes, Coordinates       */
	/*     into Output Field        */
	/********************************/
	if (FLDset_coord_units (out, units) != 1) {
		ERR_RETURN("Error setting units");
	}
	if (FLDset_nspace (out, h->nspace) != 1) {
		ERR_RETURN("Error setting nspace");
	}
	if (FLDset_nnodes (out, h->nnodes) != 1) {
		ERR_RETURN("Error setting nnodes");
	}
	if (h->nnodes)
		if (FLDset_coord(out, h->xyz,
		      h->nspace*h->nnodes, OM_SET_ARRAY_FREE) != 1) {
			ERR_RETURN("Error setting coord");
		}
	if (FLDset_ncell_sets (out, out_nsets) != 1) {
		ERR_RETURN("Error setting ncell_sets");
	}

	/********************************/
	/* Add Polyline to Output Field */
	/********************************/
	if (h->nline) {
		if (FLDadd_cell_set(out, "Polyline") != 1) {
			ERR_RETURN("Error setting line cell type");
		}
		if (FLDget_cell_set(out, set, &cell_set) != 1) {
			ERR_RETURN("Error getting line cell set");
		}
		if (FLDset_npolys(cell_set, h->nline) != 1) {
			ERR_RETURN("Error setting line ncells");
		}
		if (FLDset_poly_connect(cell_set, h->line_conn, 2*h->nline,
					OM_SET_ARRAY_FREE) != 1) {
			ERR_RETURN("Error setting line cell connect list");
		}
		set++;
	}
	else {
		ARRfree(h->line_conn);
	}

	/********************************/
	/* Add Tri Cells to Output     */
	/********************************/
	if (h->ntri) {
		if (FLDadd_cell_set(out, "Tri") != 1) {
			ERR_RETURN("Error setting Tri cell type");
		}
		if (FLDget_cell_set(out, set, &cell_set) != 1) {
			ERR_RETURN("Error getting Tri cell set");
		}
		if (FLDset_ncells(cell_set, h->ntri) != 1) {
			ERR_RETURN("Error setting Tri ncells");
		}
		if (FLDset_node_connect(cell_set, h->tri_conn, 3*h->ntri,
					OM_SET_ARRAY_FREE) != 1) {
			ERR_RETURN("Error setting Tri cell connect list");
		}
		set++;
	}
	else {
		ARRfree(h->tri_conn);
	}

	/********************************/
	/* Add Quad Cells to Output     */
	/********************************/
	if (h->nquad) {
		if (FLDadd_cell_set(out, "Quad") != 1) {
			ERR_RETURN("Error setting quad cell type");
		}
		if (FLDget_cell_set(out, set, &cell_set) != 1) {
			ERR_RETURN("Error getting quad cell set");
		}
		if (FLDset_ncells(cell_set, h->nquad) != 1) {
			ERR_RETURN("Error setting quad ncells");
		}
		if (FLDset_node_connect(cell_set, h->quad_conn, 4*h->nquad,
					OM_SET_ARRAY_FREE) != 1) {
			ERR_RETURN("Error setting quad cell connect list");
		}
		set++;
	}
	else {
		ARRfree(h->quad_conn);
	}



	/********************************************/
	/* Set Node Data on Field Output            */
	/********************************************/
	if (h->nnodes && h->data_flag ) {
		if (FLDset_node_data_ncomp (out, 1) != 1) {
			ERR_RETURN("Error setting nnode_data");
		}
		if (FLDset_node_data_comp (out,0,h->veclen,label,units) != 1) {
			ERR_RETURN("Error setting output node component");
		}

		/***
		 *** only use this for special things like texture,
		 *** or RGB, etc.
		 ***
	        data_id = 0;
		FLDset_node_data_id(out, 0, data_id);
		 **/

		if (FLDset_node_data(out, 0, (char *)h->data, h->dtype,
		     h->veclen*h->nnodes, OM_SET_ARRAY_FREE) != 1) {
			ERR_RETURN("Error setting output node data");
		}
		if (h->null_flag) {
			if (FLDset_node_null_data(out, 0,
			      (char *)&(h->null_value), h->dtype) != 1) {
			  ERR_RETURN("Error setting output null value");
			}
		}
		else {
			if (FLDset_node_null_flag(out, 0, 0) != 1) {
				ERR_RETURN("Error setting output null flag");
			}
		}
	}
	else {
		if (FLDset_node_data_ncomp (out, 0) != 1) {
			ERR_RETURN("Error setting output nnode_data");
		}
	}
	return(1);
}

/*
 * ===== REALLOC NODES  ========
 */
static int
CSbump_nodes ( OUTmesh *h,   /* internal structure storage */
               int new )     /* requested number of verts to add */
{
    int i,n;

    if (h->nnodes+new*h->nspace >= h->size)
    {
	h->size += ALLOC_SIZE;
	h->xyz = (float *)ARRrealloc(h->xyz,
			DTYPE_FLOAT, h->nspace*h->size, NULL);
        if ( h->xyz == NULL )
	    ERR_RETURN("could not realloc xyz coord array");
	if (h->data_flag)
	{
	    h->data = (char *)ARRrealloc(h->data, h->dtype,
			DTYPEtype_size[h->dtype]*h->size*h->veclen, NULL);
            if ( h->data == NULL )
		ERR_RETURN("could not realloc local data array");
        }

    }

    return(1);
}


/*
 * ===== ADD NODE DATA  ========
 */
static int
CSadd_node_data( OUTmesh *h,   /* internal structure storage */
                 int new,      /* requested number of verts to add */
                 char *ldata ) /* new node data */
{
    int i,n;
    char *p;
    int  v;

    char      *p_char, *q_char;
    int       *p_int,  *q_int;
    float     *p_flt,  *q_flt;
    short     *p_sht,  *q_sht;
    double    *p_dbl,  *q_dbl;

        v = (h->nnodes) * (h->veclen);
        p = ((char *)h->data) + DTYPEtype_size[h->dtype] * v;
	n = new * h->veclen;

	/*
	 * Set Node Data
	 */
        switch ( h->dtype )
	{
	    case DTYPE_CHAR:
	    case DTYPE_BYTE:
		q_char = (char *)ldata;
		p_char = (char *)p;
                for (i=0; i<n; i++)
                    p_char[i] = q_char[i];
                break;

	    case DTYPE_INT:
		q_int = (int *)ldata;
		p_int = (int *)p;
                for (i=0; i<n; i++)
                     p_int[i] = q_int[i];
                break;

	    case DTYPE_FLOAT:
		q_flt = (float *)ldata;
		p_flt = (float *)p;
                for (i=0; i<n; i++)
                    p_flt[i] = q_flt[i];
                break;

	    case DTYPE_SHORT:
		q_sht = (short *)ldata;
		p_sht = (short *)p;
                for (i=0; i<n; i++)
                    p_sht[i] = q_sht[i];
                break;

	    case DTYPE_DOUBLE:
		q_dbl = (double *)ldata;
		p_dbl = (double *)p;
                for (i=0; i<n; i++)
                    p_dbl[i] = q_dbl[i];
                break;

            default:
	        ERR_RETURN("data type not recognized in CSadd_node_data");
                break;
         }
	 return(1);
}

/*
 * ===== REALLOC LINE CONN  ========
 */
static int
CSbump_line_conn ( OUTmesh *h )        /* internal structure storage */
{
	static int line_conn_sz = 2;

	/*
	 * estimate connectivity array size requirements,
	 * bump temp storage if needed.
	 */
	if ((h->nline+1)*line_conn_sz >= (h->line_size)) {
		h->line_size += ALLOC_SIZE;
		h->line_conn = (int *)ARRrealloc(h->line_conn,
		    DTYPE_INT,
		    h->line_size, NULL);
	}

        if ( h->line_conn == NULL )
	  ERR_RETURN("could not realloc local line conn array");
	return(1);
}

/*
 * ===== REALLOC TRI CONN  ========
 */
static int
CSbump_tri_conn ( OUTmesh *h )        /* internal structure storage */
{
	static int tri_conn_sz = 3;
	int n;

	/*
	 * estimate connectivity array size requirements,
	 * bump temp storage if needed.
	 */
	if ((h->ntri+1)*tri_conn_sz >= (h->tri_size))
	{
		h->tri_size += ALLOC_SIZE;
		h->tri_conn = (int *)ARRrealloc(h->tri_conn,
			DTYPE_INT, h->tri_size, NULL);

            if ( h->tri_conn == NULL )
	  	ERR_RETURN("could not realloc local tri conn array");
	}

	return(1);
}

/*
 * ===== REALLOC QUAD CONN  ========
 */
static int
CSbump_quad_conn ( OUTmesh *h )        /* internal structure storage */
{
	static int quad_conn_sz = 4;
	int n;

	/*
	 * estimate connectivity array size requirements,
	 * bump temp storage if needed.
	 */
	if ((h->nquad+1)*quad_conn_sz >= (h->quad_size))
	{
		h->quad_size += ALLOC_SIZE;
		h->quad_conn = (int *)ARRrealloc(h->quad_conn,
			DTYPE_INT, h->quad_size, NULL);

            if ( h->quad_conn == NULL )
	       ERR_RETURN("could not realloc local quad conn array");
	}

	/*
         * CSprintoutstruct( h );
	 */

	return(1);
}


/*
 * ===== ADD A LINE TO THE STRUCTURE ========
 */
static int
CScreate_line ( OUTmesh *h,     /* internal structure storage */
                float *points,  /* line coord array */
                char *ldata )   /* line node data */
{
	int  i, n, nnew;

        nnew = 2;

        if (CSbump_nodes( h, nnew ) != 1)
	    ERR_RETURN("could not bump_nodes for line");

	/*
	 * Set Coordinates
	 */
        for (i=0; i < nnew*h->nspace; i++ ) {
		h->xyz[h->nnodes * h->nspace + i] = points[i];
	}

	/*
	 * Set Node Data
	 */
	if (h->data_flag) {
            if (CSadd_node_data( h, nnew, (char *)ldata ) != 1)
	        ERR_RETURN("could not add_node_data for lines");
        }

        if ( CSbump_line_conn ( h ) != 1)
	    ERR_RETURN("could not bump_line_conn");

	/*
	 * Store Connectivity for one line.
	 */
	n = h->nnodes;
	h->line_conn[nnew*h->nline  ] = n;
	h->line_conn[nnew*h->nline+1] = n+1;
	(h->nline)++;
	h->nnodes += nnew;

	return(1);
}

/*
 * ===== ADD A TRI TO THE STRUCTURE ========
 */
static int
CScreate_tri ( OUTmesh *h,     /* local storage struct */
               float *points,  /* tri corners coord array */
               char *qdata )   /* tri node data */
{
	int  i, n, nnew;

        nnew = 3;

        if (CSbump_nodes( h, nnew ) != 1)
	    ERR_RETURN("could not bump_nodes for tri");

	/*
	 * Set Coordinates
	 */
        for (i=0; i < nnew*h->nspace; i++ ) {
		h->xyz[h->nnodes * h->nspace + i] = points[i];
	}

	/*
	 * Set Node Data
	 */
	if (h->data_flag) {
            if (CSadd_node_data( h, nnew, qdata ) != 1)
	        ERR_RETURN("could not add_node_data for triangles");
	}

        if ( CSbump_tri_conn ( h ) != 1)
	    ERR_RETURN("could not bump_tri_conn");

	/*
	 * Store Connectivity for one Tri cell.
	 */
	n = h->nnodes;
	h->tri_conn[nnew*h->ntri]   = n;
	h->tri_conn[nnew*h->ntri+1] = n+1;
	h->tri_conn[nnew*h->ntri+2] = n+2;
	(h->ntri)++;
	h->nnodes += nnew;
	return(1);
}


/*
 * ===== ADD A QUAD TO THE STRUCTURE ========
 */
static int
CScreate_quad ( OUTmesh *h,     /* local storage struct */
                float *points,  /* quad corners coord array */
                char *qdata )   /* quad node data */
{
	int  i, j, n, n1, n2, nnew;

        nnew = 4;

        if (CSbump_nodes( h, nnew ) != 1)
	    ERR_RETURN("could not bump_nodes for quad");

	/*
	 * Set Coordinates
	 */
        for (i=0; i < nnew*h->nspace; i++ ) {
		h->xyz[h->nnodes * h->nspace + i] = points[i];
	}

	/*
	 * Set Node Data
	 */
	if (h->data_flag) {
            if (CSadd_node_data( h, nnew, (char *)qdata ) != 1)
	        ERR_RETURN("could not add_node_data for quads");
	}

        if ( CSbump_quad_conn ( h ) != 1)
	    ERR_RETURN("could not bump_quad_conn");

	/*
	 * Store Connectivity for one Quad cell.
	 */
	n = h->nnodes;
	h->quad_conn[nnew*h->nquad]   = n;
	h->quad_conn[nnew*h->nquad+1] = n+1;
	h->quad_conn[nnew*h->nquad+2] = n+2;
	h->quad_conn[nnew*h->nquad+3] = n+3;
	(h->nquad)++;
	h->nnodes += nnew;

	return(1);
}

/*
 * ===== ADD A POLYLINE TO THE STRUCTURE ========
 */
static int
CScreate_polyline( OUTmesh *h,     /* internal structure storage */
                   float *points,  /* line coord array */
                   char *ldata,    /* line node data */
                   int numpoints ) /* number of points in polyline array */
{
	int  i, n, nnew;

        nnew = 2;

        if (CSbump_nodes( h, numpoints ) != 1)
	    ERR_RETURN("could not bump_nodes for polyline");

	/*
	 * Set Coordinates
	 */
        for (i=0; i < numpoints*h->nspace; i++ ) {
		h->xyz[h->nnodes * h->nspace + i] = points[i];
	}

	/*
	 * Set Node Data
	 */
	if (h->data_flag) {
            if (CSadd_node_data( h, numpoints, ldata ) != 1)
	        ERR_RETURN("could not add_node_data for polylines");
        }

        if ( CSbump_line_conn ( h ) != 1)
	    ERR_RETURN("could not bump_line_conn in polyline");

	/*
	 * Store Connectivity for one line.
	 */
	n = h->nnodes;
	h->line_conn[nnew*h->nline  ] = n;
	h->line_conn[nnew*h->nline+1] = n+numpoints-1;
	(h->nline)++;
	h->nnodes += numpoints;

	return(1);
}


/*
 * ===== PRINT OUT CONTENTS OF THE STRUCTURE ========
 */

static void
CSprintoutstruct( OUTmesh *h )
{
    int i;

	printf("\ncontents of local struct:\n");
	printf(" h->nnodes    = %d\n", h->nnodes);
	printf(" h->size      = %d\n", h->size);
	printf(" h->nline     = %d\n", h->nline);
	printf(" h->line_size = %d\n", h->line_size);
	printf(" h->ntri      = %d\n", h->ntri);
	printf(" h->tri_size  = %d\n", h->tri_size);
	printf(" h->nquad     = %d\n", h->nquad);
	printf(" h->quad_size = %d\n", h->quad_size);

    for (i=0; i< h->nnodes; i++)
    {
	printf(" h->xyz[%d] = %f %f %f\n",
		 i,
		 h->xyz[3*i+0],
		 h->xyz[3*i+1],
		 h->xyz[3*i+2] );
    }
    for (i=0; i< h->ntri; i++)
    {
	printf(" h->tri_conn[%d] = %d %d %d\n",
		 i,
		 h->tri_conn[3*i+0],
		 h->tri_conn[3*i+1],
		 h->tri_conn[3*i+2] );
    }

}



/********************************
 * Resolve Parameter Block Args *
 ********************************/
static int
res_params( OMobj_id object, ParamBlockStruct *pb )
{
   OMobj_id local_id, parm_id;

   double xd;
   int i;
   char *c_ptr;

   parm_id = object;

   RESOLVE_FLT_PARM(parm_id,"zscale",OM_OBJ_RD,(pb->zscale));
   RESOLVE_FLT_PARM(parm_id,"fwidth",OM_OBJ_RD,(pb->fwidth));
   RESOLVE_FLT_PARM(parm_id,"foffset",OM_OBJ_RD,(pb->foffset));
   RESOLVE_FLT_PARM(parm_id,"threshold",OM_OBJ_RD,(pb->threshold));

   RESOLVE_INT_PARM(parm_id,"h_label_inc",OM_OBJ_RD,(pb->h_label_inc));
   RESOLVE_INT_PARM(parm_id,"draw_cityscape",OM_OBJ_RD,(pb->draw_cityscape));
   RESOLVE_INT_PARM(parm_id,"draw_sidebars",OM_OBJ_RD,(pb->draw_sidebars));
   RESOLVE_INT_PARM(parm_id,"draw_ticmarks",OM_OBJ_RD,(pb->draw_ticmarks));
   RESOLVE_INT_PARM(parm_id,"draw_labels",OM_OBJ_RD,(pb->draw_labels));
   RESOLVE_INT_PARM(parm_id,"draw_threshold",OM_OBJ_RD,(pb->draw_threshold));

   RESOLVE_INT_PARM(parm_id,"thresh_choice",OM_OBJ_RD,(pb->thresh_choice));
   RESOLVE_INT_PARM(parm_id,"hollow_mode",OM_OBJ_RD,(pb->hollow_mode));
   RESOLVE_INT_PARM(parm_id,"auto_norm",OM_OBJ_RD,(pb->auto_norm));

   return(1);
}


/***********************************
 * Create A Generic Box as 6 Quads *
 * with shared corner nodes        *
 ***********************************/
static int
CScreate_box ( ParamBlockStruct *pb,  /* input parameter structure storage */
            OUTmesh *h,          /* output structs */
            float xpos,       /* positions of base */
            float ypos,
            float height,     /* height of block */
            float width,      /* width, away from center */
            float data_val,   /* node data value */
            int hilite )    /* any special value to attatch? */

{
	int  i, j, k, n, nnew;
        int  numpoints;
        float ldata[8];

        /*
         * define the topology of a reference box made of quads
         */
        static int box_topo[6][4] = {
            {0, 1, 2, 3},
            {4, 5, 6, 7},
            {1, 2, 6, 5},
            {0, 1, 5, 4},
            {0, 4, 7, 3},
            {3, 7, 6, 2},
        };

        static float box_verts[8][3] = {
          /* x,  y,  z */
            {0., 0., 0.},
            {1., 0., 0.},
            {1., 1., 0.},
            {0., 1., 0.},
            {0., 0., 1.},
            {1., 0., 1.},
            {1., 1., 1.},
            {0., 1., 1.},
        };

        float aa[3], bb[3], cc[3];

        /**************
         * Body       *
         **************/

        nnew = 4;
        numpoints = 8;

        if (CSbump_nodes( h, numpoints ) != 1)
	    ERR_RETURN("could not bump_nodes for create box");

	/*
	 * Set Coordinates
	 */
        aa[0] = -0.5;  aa[1] = -0.5;   aa[2] = 0.0;
        bb[0] = width; bb[1] = width;  bb[2] = height;
        cc[0] = xpos;  cc[1] = ypos;   cc[2] = 0.0;
        k=0;
        for (i=0; i < numpoints; i++ )
        for (j=0; j < h->nspace; j++ )
	{
		h->xyz[h->nnodes * h->nspace + k++] =
                     (box_verts[i][j] + aa[j])*bb[j]+cc[j];
        }

	/*
	 * Set Node Data
	 */

        /**
         ** Assumes FLOAT!
         **/
	if ( hilite )
	{
           for (i=0; i < numpoints; i++ )
             ldata[i] = 0.0;
        }
	else
	{
           for (i=0; i < numpoints; i++ )
             ldata[i] = (float) data_val;
        }

	if (h->data_flag)
	{
            if (CSadd_node_data( h, numpoints, (char *)ldata ) != 1)
	        ERR_RETURN("could not add_node_data for box");
        }

	/*
	 * Store Connectivity for 6 quads.
	 */
        n = h->nnodes;
        for (i=0;i<6;i++) {

            if ( CSbump_quad_conn ( h ) != 1)
	        ERR_RETURN("could not bump_quad_conn in create_box");

	    h->quad_conn[nnew*(h->nquad)]   = n + box_topo[i][0];
	    h->quad_conn[nnew*(h->nquad)+1] = n + box_topo[i][1];
	    h->quad_conn[nnew*(h->nquad)+2] = n + box_topo[i][2];
	    h->quad_conn[nnew*(h->nquad)+3] = n + box_topo[i][3];
	    (h->nquad)++;
        }
        h->nnodes += numpoints;

        return(1);
}

/**************************************
 * Create A Generic Box in Wireframe  *
 **************************************/
static int
draw_wire_box ( ParamBlockStruct *pb,  /* input parameter structure storage */
            OUTmesh *h,       /* output structs */
            float xpos,       /* positions of base */
            float ypos,
            float height,     /* height of block */
            float width,      /* width, away from center */
            float data_val,   /* node data value */
            int hilite )      /* any special value to attatch? */
{
	int  i, j, k, n, nnew;
        int  numpoints;
        float ldata[16];
	float points[48];
	float points2[6];

        static float box_verts[10][3] = {
          /* x,  y,  z */
            {0., 0., 0.},
            {1., 0., 0.},
            {1., 1., 0.},
            {0., 1., 0.},
            {0., 0., 0.},
            {0., 0., 1.},
            {1., 0., 1.},
            {1., 1., 1.},
            {0., 1., 1.},
            {0., 0., 1.},
        };

        float aa[3], bb[3], cc[3];

        /**************
         * Body       *
         **************/

        numpoints = 10;

	/*
	 * Set Coordinates
	 */
        aa[0] = -0.5;  aa[1] = -0.5;   aa[2] = 0.0;
        bb[0] = width; bb[1] = width;  bb[2] = height;
        cc[0] = xpos;  cc[1] = ypos;   cc[2] = 0.0;
        k=0;
        for (i=0; i < numpoints; i++ )
        for (j=0; j < h->nspace; j++ )
	{
		points[k++] = (box_verts[i][j] + aa[j])*bb[j]+cc[j];
        }

	/*
	 * Set Node Data
	 */
	if ( hilite )
	{
           for (i=0; i < numpoints; i++ )
             ldata[i] = 0.0;
        }
	else
	{
           for (i=0; i < numpoints; i++ )
             ldata[i] = (float) data_val;
        }

        if (CScreate_polyline( h, points, (char *)ldata, 10) != 1)
		     ERR_RETURN("error in CScreate_polyline");

        for (i=0; i < 3; i++ )
	{
            points2[3*0+i] = points[3*1+i];
            points2[3*1+i] = points[3*6+i];
	}
        if (CScreate_line( h, points2, (char *)ldata ) != 1)
		     ERR_RETURN("error in CScreate_line");

        for (i=0; i < 3; i++ )
	{
            points2[3*0+i] = points[3*2+i];
            points2[3*1+i] = points[3*7+i];
	}
        if (CScreate_line( h, points2, (char *)ldata ) != 1)
		     ERR_RETURN("error in CScreate_line");

        for (i=0; i < 3; i++ )
	{
            points2[3*0+i] = points[3*3+i];
            points2[3*1+i] = points[3*8+i];
	}
        if (CScreate_line( h, points2, (char *)ldata ) != 1)
		     ERR_RETURN("error in CScreate_line");


        return(1);
}


/**********************************************************************/
/* Primary City Scape Geometric Object Creation Routine               */
/*   first scalar controls height, 2nd if present, controls colour    */
/**********************************************************************/

static int
draw_city( ParamBlockStruct *pb, OUTmesh *h, int nx, int ny,
           float **values, float **values2,
           int veclen, int ncomp, int h_flag )
{
   int i,j;
   int index;
   float xpos, ypos, height;
   float spacing;
   int hilite;
   float val, val2;
   int   ix, iy;
   float w;
   float *fp;
   float largest_dim;

   /* preset highlite to not work for now */
   ix = iy = -1;

   if ( nx > ny )
       largest_dim = (float)nx;
   else
       largest_dim = (float)ny;

   /*
    * Loop over all rows and columns of 2D array
    */
   for (j=0; j<ny; j++) {
     for (i=0; i<nx; i++) {

       /*
	* set the visible attributes of box
	*/
       if ( ncomp > 1 && veclen == 1 )
       {
	   fp  = values[j];
           val = fp[i];
	   if ( values2 )
	   {
	       fp  = values2[j];
               val2 = fp[i];
	   }
	   else
	       val2 = val;
       }
       else
       {
	   fp  = values[0];
           val = fp[(j*nx+i)*veclen];
	   if ( values2 )
	   {
	       fp  = values2[0];
               val2 = fp[(j*nx+i)*veclen];
	   }
	   else
	       val2 = val;
       }

       xpos  = (float) i;
       ypos  = (float)(ny-j-1);

       if ( pb->auto_norm )
       {
           height = pb->zscale * (val/pb->max) * largest_dim;
       }
       else
       {
           height = pb->zscale * val;
       }

       /*
        * Switch if it has a picked hilite or not
        */
       if ( i==ix && j==iy )
               hilite=1;
       else
               hilite=0;

       /*
	* Threshold Sheet Selection, for NULL data
	*/
       if ( (pb->thresh_choice == 0) ||                           /* 0=all */
	    (pb->thresh_choice == 1 && val <= pb->threshold) ||   /* 1=below */
	    (pb->thresh_choice == 2 && val >  pb->threshold) )    /* 2=above*/
       {
	  if ( h_flag == 0 )
	  {
               if (CScreate_box( pb, h, xpos, ypos, height,
                           pb->fwidth, val2, hilite ) != 1)
		     ERR_RETURN("error in draw_a_box");
          }
       }
       else if ( pb->hollow_mode )
       {
          /* Draw the cells in hollow wireframe mode */
	  if ( h_flag == 1 )
	  {
               if (draw_wire_box( pb, h,  xpos, ypos, height,
                           pb->fwidth, val2, hilite ) != 1)
		     ERR_RETURN("error in draw_a_box");
          }
       }
     }
   }
   return(1);
}

/**********************************************************************/
/* Draw the line annotation decoration around the city  */
/**********************************************************************/
static int
draw_base_line ( ParamBlockStruct *pb, OUTmesh *h, int nx, int ny )
{
    float data_val;
    int i,j;
    float xs, ys, xe, ye, zs, ze;
    float minx, miny, maxx, maxy;
    float points[15];
    float ldata[5];

    minx = -0.5 * pb->fwidth;
    miny = -0.5 * pb->fwidth;
    maxx = ((float)(nx-1) + pb->fwidth * 0.5);
    maxy = ((float)(ny-1) + pb->fwidth * 0.5);
    data_val = 0.0;
    ldata[0] = ldata[1] = ldata[2] = ldata[3] = ldata[4] = data_val;

    /*
     *  Draw 4 lines to make a polygon frame around the base
     *  at the city limits, needs 5 points
     */
    points[0]=minx;   points[1]=miny;  points[2]=0.0;
    points[3]=maxx;   points[4]=miny;  points[5]=0.0;
    points[6]=maxx;   points[7]=maxy;  points[8]=0.0;
    points[9]=minx;   points[10]=maxy; points[11]=0.0;
    points[12]=minx;  points[13]=miny; points[14]=0.0;
    if (CScreate_polyline( h, points, (char *)ldata, 5) != 1)
		     ERR_RETURN("error in CScreate_polyline");

    /*
     * Draw small tic lines out from base
     */
    for (i=0; i<ny; i++) {
        points[0]=minx;
        points[1]=(float)(ny-i-1);
        points[2]=0.0;
        points[3]=minx - ( 0.10 * (float)nx);
        points[4]=(float)(ny-i-1);
        points[5]=0.0;

        if (CScreate_line( h, points, (char *)ldata ) != 1)
		     ERR_RETURN("error in CScreate_line");

    }
    for (j=0; j<nx; j++) {

        points[0]=(float)(j);
        points[1]=miny;
        points[2]=0.0;
        points[3]=(float)(j);
        points[4]=miny - ( 0.10 * (float)ny);
        points[5]=0.0;

        if (CScreate_line( h, points, (char *)ldata ) != 1)
		     ERR_RETURN("error in CScreate_line");

    }

    return(1);
}

/**********************************************************************/
/* Draw the Threshold Sheet  */
/**********************************************************************/
static int
draw_threshold ( ParamBlockStruct *pb, OUTmesh *h, int nx, int ny )
{
    int i,j;
    float minx, miny, maxx, maxy;
    float points[12];
    float ldata[4];
    float height;

    height = pb->zscale * pb->threshold;
    minx = -0.5 * pb->fwidth;
    miny = -0.5 * pb->fwidth;
    maxx = ((float)(nx-1) + pb->fwidth * 0.5);
    maxy = ((float)(ny-1) + pb->fwidth * 0.5);
    ldata[0] = ldata[1] = ldata[2] = ldata[3] = pb->threshold;

    /*
     *  quadrilateral frame over the base
     *  at the city limits
     */
    points[0]=minx;   points[1]=miny;  points[2]=height;
    points[3]=maxx;   points[4]=miny;  points[5]=height;
    points[6]=maxx;   points[7]=maxy;  points[8]=height;
    points[9]=minx;   points[10]=maxy; points[11]=height;
    if (CScreate_quad( h, points, (char *)ldata ) != 1)
		     ERR_RETURN("error in CScreate_polyline");

    return(1);
}


/**********************************************************************/
/* Secondary Side Bar Geometric Object Creation Routine */
/**********************************************************************/
static int
draw_averages( ParamBlockStruct *pb, OUTmesh *h, int nx, int ny,
               float **values, int veclen, int ncomp )
{
   int i,j;
   int index;
   float xpos, ypos, height;
   float base;
   double average, minimum, maximum, val;
   float data;
   float norm_nx, norm_ny;
   float points[12];
   float ldata[4];
   float w;
   float *fp;
   float largest_dim;

   norm_nx = (float) nx;
   norm_ny = (float) ny;
   w = pb->fwidth;

   if ( nx > ny )
       largest_dim = (float)nx;
   else
       largest_dim = (float)ny;

   for (i=j=0; j<ny; j++)
   {
       average = 0.0;
       for (i=0; i<nx; i++)
       {
           if ( ncomp > 1 )
           {
	       fp  = values[j];
               val = fp[i*veclen];
           }
           else
           {
	       fp  = values[0];
               val = fp[(j*nx+i)*veclen];
           }

	   if ( i==0 ) minimum = maximum = val;
           if ( maximum < val ) maximum = val;
           if ( minimum > val ) minimum = val;
           average += val;
       }
       average = average/norm_nx;

       xpos  = (float)nx * ( 1.0 + pb->foffset);
       ypos  = (float)(ny-j-1);

       if ( pb->auto_norm )
       {
           height = pb->zscale * (average/pb->max) * largest_dim;
           base   = pb->zscale * (minimum/pb->max) * largest_dim;
       }
       else
       {
           height = pb->zscale * average;
           base   = pb->zscale * minimum;
       }

       points[0]=xpos;   points[1]=(-0.5*w+ypos);  points[2]=base;
       points[3]=xpos;   points[4]=(-0.5*w+ypos);  points[5]=height;
       points[6]=xpos;   points[7]=(+0.5*w+ypos);  points[8]=height;
       points[9]=xpos;   points[10]=(+0.5*w+ypos); points[11]=base;

       ldata[0]=ldata[1]=ldata[2]=ldata[3]= minimum;
       if ( CScreate_quad  ( h, points, (char *)ldata ) != 1)
		     ERR_RETURN("error in CScreate_quad");

       if ( pb->auto_norm )
       {
           height = pb->zscale * (maximum/pb->max) * largest_dim;
           base   = pb->zscale * (average/pb->max) * largest_dim;
       }
       else
       {
           height = pb->zscale * maximum;
           base   = pb->zscale * average;
       }

       points[0]=xpos;   points[1]=(-0.5*w+ypos);  points[2]=base;
       points[3]=xpos;   points[4]=(-0.5*w+ypos);  points[5]=height;
       points[6]=xpos;   points[7]=(+0.5*w+ypos);  points[8]=height;
       points[9]=xpos;   points[10]=(+0.5*w+ypos); points[11]=base;

       ldata[0]=ldata[1]=ldata[2]=ldata[3]=  maximum;
       if ( CScreate_quad  ( h, points, (char *)ldata ) != 1)
		     ERR_RETURN("error in CScreate_quad");
   }

   for (i=j=0; i<nx; i++)
   {
       average = 0.0;
       for (j=0; j<ny; j++)
       {
           if ( ncomp > 1 )
           {
	       fp  = values[j];
               val = fp[i*veclen];
           }
           else
           {
	       fp  = values[0];
               val = fp[(j*nx+i)*veclen];
           }

	   if ( j==0 ) minimum = maximum = val;
           if ( maximum < val ) maximum = val;
           if ( minimum > val ) minimum = val;
           average += val;
       }
       average = average/norm_ny;

       xpos  = (float) i;
       ypos  = (float) ny * (1.0 + pb->foffset);

       if ( pb->auto_norm )
       {
           height = pb->zscale * (average/pb->max) * largest_dim;
           base   = pb->zscale * (minimum/pb->max) * largest_dim;
       }
       else
       {
           height = pb->zscale * average;
           base   = pb->zscale * minimum;
       }

       points[0]=(-0.5*w+xpos);   points[1]=ypos;  points[2]=base;
       points[3]=(-0.5*w+xpos);   points[4]=ypos;  points[5]=height;
       points[6]=(+0.5*w+xpos);   points[7]=ypos;  points[8]=height;
       points[9]=(+0.5*w+xpos);   points[10]=ypos; points[11]=base;

       ldata[0]=ldata[1]=ldata[2]=ldata[3]= minimum;
       if ( CScreate_quad  ( h, points, (char *)ldata ) != 1)
		     ERR_RETURN("error in CScreate_quad");

       if ( pb->auto_norm )
       {
           height = pb->zscale * (maximum/pb->max) * largest_dim;
           base   = pb->zscale * (average/pb->max) * largest_dim;
       }
       else
       {
           height = pb->zscale * maximum;
           base   = pb->zscale * average;
       }

       points[0]=(-0.5*w+xpos);   points[1]=ypos;  points[2]=base;
       points[3]=(-0.5*w+xpos);   points[4]=ypos;  points[5]=height;
       points[6]=(+0.5*w+xpos);   points[7]=ypos;  points[8]=height;
       points[9]=(+0.5*w+xpos);   points[10]=ypos; points[11]=base;

       ldata[0]=ldata[1]=ldata[2]=ldata[3]= maximum;
       if ( CScreate_quad  ( h, points, (char *)ldata ) != 1)
		     ERR_RETURN("error in CScreate_quad");

   }

   return(1);
}

